#parameter specification

# "nchain" is how many chains you want to use when performing the RCEM
# "nclust" is the number of clusters (increment this by one until BIC goes up)
# "threshold" is the p-value cut-off for the RCEM algorithm
# replace "data.txt" with your data input file with the following tab-delimited
# format:
#
# genename     time1  time2  time3 ...
# geneA 0.54          0.22  ...
# geneB 0.11   0.76   0.004 ...
# ...

## step I ##

all = function(nclust,threshold,nchain){
 tol = 1e-5
 my.data=read.table("C:\\..\\gene.txt", header=T, na.strings =" ", row.names=1,sep="\t")

iter.max = 500
#end parameter specification
#___________________________________________________________________________________



#_____________________

my.name=rownames(my.data)
 my.sample=data.matrix(my.data)
n=dim(my.sample)[1]
m=dim(my.sample)[2]
if( n/m < 5) print("The dataset is too small to have so many clusters")
tt=1:m
library(gss)
source("C:\\..\\ssanova1.R")
source("C:\\..\\mcemBIC.R")
like.rec=bic.rec=NULL

my.clust=list()
clust.centers=list()
for(chain in 1:nchain){ #chain = 1
print(paste("chain", chain, "is running"), quote = F)
my.sample.vec=as.numeric(my.sample)
my.which=which(is.na(my.sample.vec))
my.sample.vec[my.which]=0
my.sample0=matrix(my.sample.vec, n, m)
clust=kmeans(my.sample0,nclust)$cluster
zeta=NULL
varht=NULL
clust.initial=matrix(nrow=nclust,ncol=m)

for(kk in 1:nclust){ ## kk=59
         label=which(clust==kk)
         temp=NULL
         temp.pred=NULL
         clusti.ini=as.numeric(as.vector(t(my.sample[label,])))
         tm.ini=rep(tt,length(label))
         my.geneid.ini=as.factor(sort(rep(1:length(label),m)))
         my.which=which(as.numeric(is.na(clusti.ini))==0)
         clusti=clusti.ini[my.which]
         tm=tm.ini[my.which]
         my.geneid=my.geneid.ini[my.which]
         formula=~1|my.geneid
         form.wk <- terms.formula(formula)[[2]]
         term.wk <- strsplit(deparse(form.wk), " \\| ")[[1]]
         z2.wk <- eval(parse(text = term.wk[2]))
         z <- NULL
         lvl.z2 <- levels(z2.wk)
         for (k in lvl.z2) z <- cbind(z, as.numeric(z2.wk == k))
         init <- -1
         env <- length(levels(z2.wk))
         fun <- function(zeta, env) diag(10^(-zeta), env)
         sigma <- list(fun = fun, env = env)    	
         my.ran=list(z = z, sigma = sigma, init = init)
         par <- list(nphi=1,mkphi=mkphi.cubic,mkrk=mkrk.cubic,
            env=c(min(tm),max(tm))+c(-1,1)*(max(tm)-min(tm))*.00)

         temp=ssanova(clusti~tm,
                random=my.ran,type=list(tm=list("custom",par)),  id.basis=1:m, alpha=1)
         temp.pred=predict(temp,data.frame(tm=tt,offset=rep(0,length(tt))))
         clust.initial[kk,]=temp.pred
         my.geneid=as.factor(sort(rep(1:length(label),m)))
         formula=~1|my.geneid
         form.wk <- terms.formula(formula)[[2]]
         term.wk <- strsplit(deparse(form.wk), " \\| ")[[1]]
         z2.wk <- eval(parse(text = term.wk[2]))
         z <- NULL
         lvl.z2 <- levels(z2.wk)
         for (k in lvl.z2) z <- cbind(z, as.numeric(z2.wk == k))
         init <- -1
         env <- length(levels(z2.wk))
         fun <- function(zeta, env) diag(10^(-zeta), env)
         sigma <- list(fun = fun, env = env)    	
         my.ran=list(z = z, sigma = sigma, init = init)
         random.zb=my.ran$z%*%temp$b
         temp.pred=predict(temp,data.frame(tm=tt))
         zeta[kk]=temp$zeta
         varht[kk]=temp$varht
         clust.initial[kk,]=temp.pred
         rm(temp)
         print(paste("cluster",kk, "is calculated"), quote = F)
 }

clust.sim=NULL
clust.sim=em.clust(my.sample,tt,clust,nclust,clust.initial,zeta,varht,thres=threshold,iter.max,tol)#
clust.centers[[chain]]=clust.sim$mean.curve
my.zeta<-my.varht<-gamma.prior<-NULL
my.zeta=clust.sim$var.ratio
my.varht=clust.sim$variance
gamma.prior=clust.sim$pp
like.rec[chain]=clust.sim$likelihood
bic.rec[chain]=clust.sim$bic
clust.temp=NULL
for(i in 1:n)
	clust.temp[i]=post(my.sample[i,],nclust,gamma.prior,clust.centers[[chain]],my.zeta,my.varht)$clust
my.clust[[chain]]=clust.temp
}

####prepare output

output.clust=NULL


output.clust=my.clust[[which(bic.rec==min(bic.rec))[1]]]
output.center=matrix(nrow=nclust,ncol=m)
output.center=clust.centers[[which(bic.rec==min(bic.rec))[1]]]
output.bic=min(bic.rec)

dump(c("output.clust","output.center","output.bic"),"output.Rdata")

#return( list( output.bic = output.bic , nclust = nclust ) )

#}

num.k1 = 2
num.k2 = 50
threshold = 0.1
nchain = 1

num.clust.bic = matrix( 0, nrow = (num.k2-num.k1+1), ncol = 2 )

for ( num.clust in num.k1  : num.k2 ) {
  all.result = all( num.clust,threshold,nchain )
  num.clust.bic[ (num.clust - num.k1 + 1), 1 ] = all.result$nclust
  num.clust.bic[ (num.clust - num.k1 + 1), 2] = round(all.result$output.bic,2)
}
