
### step iii identification ###

library(splines)
library(MASS)
library(mvtnorm)


hdscad <- function(Xorg, yorg, betaint, lambda, a=3.7, stand.mean=T, stand.scale=F, eps=1e-5, trace=F,select.method)
{
    large =1000000*(lambda+1)
    ndim = ncol(Xorg)

    nobs = nrow(Xorg)
    beta = rep(0,ndim)
    ### normalization
    muy = mean(yorg)
    y   = yorg-muy
    ####################################################
    ### Center x and scale x, and save the means and sds
    if(stand.mean == T){
        mu = apply(Xorg, 2, mean)
        X  = sweep(Xorg, 2, mu, "-")   # centers x
    }
    if(stand.scale ==T){
        sdx = apply(Xorg, 2, sd)
        X   = sweep(Xorg, 2, sdx, "/")    # scales x
    }
    ####################################################
    boolcccp = 1
    betac = betaint
    dd    = c(1:ndim)

    ###################
    conv = c()
    iter = 1
    inner.iter = 0
    while(boolcccp==1){
        JD         = J1(betac,lambda,a)
        beta       = rep(0,ndim)
        residual   = y-X%*%beta
        gradient   = -t(X)%*%residual/nobs+JD
        gmax       = max(abs(gradient))
        po         = which.max(abs(gradient))
        gamma      = rep(0,ndim)
        gamma[po]  = - sgn(gradient[po])
        active     = rep(0,ndim)
        active[po] = 1

        boollasso=1
        if(abs(gradient[po])<lambda)
            boollasso = 0

        while(boollasso==1){
            inner.iter <- inner.iter + 1
            numer = t(X) %*% X%*% gamma/nobs
            agradient=gradient[po]
            anumer=numer[po]

            d1 =(agradient-gradient)/(numer-anumer)
            d2 =-(agradient+gradient)/(numer+anumer)
            d1[d1<=0] = large
            d2[d2<=0] = large
            d = pmin(d1,d2)
            d[active==1] = -beta[active==1]/gamma[active==1]
            d[(active==1) & (d<=0)] = large
            po = which.min(d)

            if(d[po]>large-eps){
                beta      = beta+abs(gradient)*gamma
                gradient  = -t(X)%*%(y-X%*%beta)/nobs+JD
                boollasso = 0
            }else{
                add      = 0
                beta     = beta+d[po]*gamma
                gradient = -t(X)%*%(y-X%*%beta)/nobs+JD

                if(max(abs(gradient)) <= lambda){
                    potmp = min(dd[active==1])
                    if(abs(gradient[potmp])<lambda){
                        eta=lambda-abs(gradient[potmp])
                        beta=beta-eta*gamma
                        boollasso=0
                        gradient=-t(X)%*%(y-X%*%beta)/nobs+JD
                    }
                    }else{
                    if(active[po] == 1)
                        active[po]=0
                    else{
                        active[po]=1
                        add<-1
                        poadd<-po
                    }
                    if(max(abs(gradient))< eps)
                        boollasso <- 0
                    po=min(dd[active==1])
                    XA=X[,active==1]
                    C = t(XA)%*%XA/nobs
                    if(det(C) < 1e-5){
                        boollasso <- 0
                        cat("Selected variables are strongly correlated!!!\n")
                    }
                    gamma<-sgn(beta)

                    if(add==1){
                        gamma[poadd]<- -sgn(gradient[poadd])
                    }
                    gammaA=ginv(C)%*%gamma[active==1]
                    gamma[active==1]<-gammaA
                    gamma[active==0]<-0
                    }
                }
        } #end of lasso
        #cat("inner iteration", inner.iter, "\n")

        # deviance
        optc = 1/2/nobs*sum((y-X%*%betac)^2) + Jlambda(betac,lambda,a)
        opt  = 1/2/nobs*sum((y-X%*%beta)^2)  + Jlambda(beta,lambda,a)
        conv[iter] <- optc
        if(abs(optc-opt)<eps){
            if(trace==T)
                cat("Convergence of optimization !!\n")
            iter = iter+1
            conv[iter] = opt
            boolcccp=0
        }else if(sum(abs(J1(beta,lambda,a)-JD))<eps){
            if(trace==T)
                cat("Convergence of penalty gradient !!\n")
            iter = iter+1
            conv[iter] = opt
            boolcccp=0
        }else if( sum(abs(beta-betac)) < eps){
            if(trace==T)
                cat("Convergence of coefficients !!\n")
            iter = iter+1
            conv[iter] = opt
            boolcccp=0
        }else if(optc+eps < opt & iter >1){
            if(trace==T)
                cat("Convergence 00000 !!\n")
            beta = betac
            opt  = optc
            boolcccp = 0
        }
        else{
            iter <- iter + 1
            betac =beta
        }
    } #End of CCCP
    ERROR = mean((y-X%*%beta)^2)

    if(select.method == 'GCV') {
    #### GCV
  print('yes!!!')
    dd = c(1:ndim)[abs(beta)>0]
    if(length(dd)!=0){
        XA  = X[,dd]
        tXA = t(XA)
        CA  = tXA %*% XA

        Sigma = rep(0,ndim)
        p1    = abs(J1(beta,lambda,a)+lambda*sgn(beta))
        Sigma[abs(beta)>0] = p1[abs(beta)>0]/abs(beta)[abs(beta)>0]
        Sigma              = diag(Sigma)
        PX = XA %*% ginv(CA +nobs*Sigma[dd]) %*% tXA
        GCV = mean((y-X%*%beta)^2)/(1-sum(diag(PX))/nobs)^2
        ERROR = mean((y-X%*%beta)^2)
        ETHETA = sum(diag(PX))/nobs
    }
    else{
        ERROR  = mean((y^2))
        ETHETA = ndim/nobs
        GCV = mean((y^2))
    }
} else {

     dd = c(1:ndim)[abs(beta)>0]
    if(length(dd)!=0){
        XA  = X[,dd]
        tXA = t(XA)
        CA  = tXA %*% XA

        Sigma = rep(0,ndim)
        p1    = abs(J1(beta,lambda,a)+lambda*sgn(beta))
        Sigma[abs(beta)>0] = p1[abs(beta)>0]/abs(beta)[abs(beta)>0]
        Sigma              = diag(Sigma)
        PX = XA %*% ginv(CA +nobs*Sigma[dd]) %*% tXA
     ####   GCV = mean((y-X%*%beta)^2)/(1-sum(diag(PX))/nobs)^2
        ERROR = mean((y-X%*%beta)^2)
        ETHETA = sum(diag(PX))/nobs
       GCV = log(ERROR) + ETHETA * log(nobs) / nobs
    }
    else{
        ERROR  = mean((y^2))
        ETHETA = ndim/nobs
        GCV = log(ERROR)
    }


}
    #### Variance
    is.corr = 0
    Varbeta = rep(0,ndim)
    if(length(dd)!=0){
        if(det(CA/nobs) <1e-5){
            is.corr = 1
        }
        sigma2 = (y-X%*%beta)^2
        sigma2 = diag(drop(sigma2))
        covl = tXA%*%sigma2%*%XA/nobs-(tXA%*%(y-X%*%beta)/nobs)%*%t(tXA%*%(y-X%*%beta)/nobs)
        p1 = abs((J1(beta, lambda, a)+lambda*sgn(beta))[dd])/abs(beta)[dd]
        inv.CA_lambda = ginv(CA + nobs*diag(p1,length(dd),length(dd)))
        Varbeta[dd]   = diag(inv.CA_lambda %*% covl %*% inv.CA_lambda)
        Varbeta[dd]   = ifelse(Varbeta[dd] <1e-10, 0, Varbeta[dd])
        Varbeta[dd]   = sqrt(nobs*Varbeta[dd])
    }
    if(stand.scale==F){
        sdx = rep(1, ndim)
    }
    if(stand.mean==T){
        beta0 = muy-sum(beta/sdx*mu)
        beta  = beta/sdx
    }else{
        beta0 = muy
        beta  = beta/sdx
    }
    return(list(coefficients = c(beta0, beta),
                         std = Varbeta,
                      lambda = lambda,
                           a = a,
                         GCV = GCV,
                       meanx = mu,
                       meany = muy,
                    gradient = drop(gradient),
               fitted.values = drop(X %*% beta+beta0),
                 convergence = conv,
                        type = "REGRESSION",
                       mtype = "hdscad",
                     is.corr = is.corr,
                       ERROR = ERROR,
                      ETHETA = ETHETA,
                       iter  = iter,
                     nonzero = length(dd),
                   objective = opt)
          )

}

sgn <- function(x){
    ndim<-length(x)
    sgn<-rep(0,ndim)
    sgn[x>0]<-1
    sgn[x<0]<- -1
    return(sgn)
}

J1<-function(beta,lambda,a){
    ndim  <- length(beta)
    abeta <- abs(beta)
    J1    <- rep(0,ndim)
    J1[(abeta> a*lambda)]<- -lambda
    dd                   <- c(1:ndim)[(lambda< abeta) & (abeta<= a*lambda)]
    J1[dd]               <- (a*lambda-abeta[dd])/(a-1)-lambda
    J1[beta<0]=-J1[beta<0]
    return(J1)
}
Jlambda <- function(beta,lambda,a)
{

    ndim  <- length(beta)
    abeta <- abs(beta)
    J <- rep(0,ndim)
    J[abeta<lambda] <- lambda*abeta[abeta <lambda]
    dd              <- c(1:ndim)[(lambda< abeta) & (abeta<= a*lambda)]
    J[dd]           <- (a*lambda*(abeta[dd]-lambda)-(abeta[dd]^2-lambda^2)/2)/(a-1) + lambda^2
    J[(abeta> a*lambda)]<- (a-1)*lambda^2/2 + lambda^2
    return(sum(J))
}


secstep = function(lambda0,lambda1,step,curv,tt.grid,nclust) {


FrstD =  matrix( nrow = length(tt.grid), ncol = nclust )
x.curv = t(curv)
betaint = rep( 0, nclust )
coef = matrix(nrow = nclust, ncol = nclust+1 )

postscript("C:\\..\\BIC.ps")
par(mfrow=c(3,3))
  for ( i in 1 : nclust )
{
inter = interpSpline( curv[i,] ~ tt.grid )

FrstD[,i] = predict( inter , tt.grid, deriv = 1)$y
gcv.min = 10000000
trk = 0
num = (lambda1 - lambda0)/step
gcv.trk = rep(0,num+1)
x = seq( lambda0, lambda1, by = step )
  for ( lambda in x )
    { trk = trk + 1
      hdscad0 = hdscad( Xorg = x.curv, yorg = FrstD[,i], betaint = betaint, lambda = lambda,stand.scale=T,
                select.method = 'BIC' )
      gcv = hdscad0$GCV
      gcv.trk[trk] = gcv
      if ( gcv < gcv.min )
        {
          gcv.min = gcv
          lambda2 = hdscad0$lambda
        }
    }
 hdscad1 = hdscad( Xorg = x.curv, yorg = FrstD[,i], betaint = betaint,lambda = lambda2,select.method = 'BIC' )
 coef[i,] = hdscad1$coefficients

plot(x, gcv.trk, xlab = 'lamdba', ylab= 'GCV',main=paste('cluster',i,sep=""))
}
dev.off()

write.csv(coef,file="C:\\..\\coef.csv")
write.table(coef, file="C:\\..\\coef.dat",sep="  ", col.names=F, row.names=F)

return(coef=coef)
}

    lambda0 = 0.0001
    lambda1 = 0.05
    step = 0.0001

tt.grid = ( 1:dim(curv)[2] -1 ) * 17 / ( dim(curv)[2] -1 )
secout = secstep(lambda0=lambda0,lambda1=lambda1,step=step,curv=curv,tt.grid=tt.grid,nclust=41)
coef = secout
